function [ ground_points_bool ] = PMF( pc, do_global_thresholding, exponential_, w_max,  dh_0, dh_max, b, c, s)
% 
% 
% 
% This function is an implementation of [1]. It segments an urban/village pointcloud to ground and
% non-ground points using the height-map morphological-operations based on the method from [1].
%
%   Required Inputs:
%   Data:
%   pc - an urban\rural pointCloud object
%   
%   Parameters:
%   Parameter names are generally coherent with [1]. Corresponding names from PCL code are mentioned in comments.
%   w_max %PCL: max_window_size 4
%   dh_0 %PCL: initial_distance_ 7
%   dh_max %PCL: max_distance_ 8
%   b %PCL: base_ 3
%   exponential_ %PCL: exponential_ 5
%   c %PCL: cell_size_ 2
%   s %PCL: slope_ 6
%   do_global_thresholding %hard coded as true in PCL
%
%   Outputs:
%   ground_points_bool - a boolean indicator vector. indicates points
%   that are ground points
%
%   References-
%  [1] Zhang, K., Chen, S. C., Whitman, D., Shyu, M. L., Yan, J., & Zhang, C. (2003). 
%      A progressive morphological filter for removing nonground measurements from airborne LIDAR data. 
% 	   IEEE transactions on geoscience and remote sensing, 41(4), 872-882.
% 



%% create minimum-interpolation-2D-image
    
% number of boxes in each dimension
boxNum = @(vec, c) 1+floor((max(vec)-min(vec))/c);
m = boxNum(pc.Location(:,2), c);
n = boxNum(pc.Location(:,1), c);

[XY_indices,gridSize,~,~] = createOrthogonalGrid(pc, c);
z = pc.Location(:,3);

A = accumarray([double(XY_indices(:,2)) double(XY_indices(:,1))],double(z),fliplr(gridSize),@min,inf);

% IMPORTANT MATRIX!! stores which pixel is associated to points from cloud (and which isn't)
BW = not(isinf(A)); 
%     ignore_mask = not(imdilate(BW, strel('disk',2)));
[~, IDX] = bwdist(BW);

A=A(:);
IDX=IDX(:);
A=A(IDX);
A=reshape(A,m,n);
    
%% calculate w (PCL: window_sizes)
if exponential_
    b_to_the_iter_num = (w_max)/2;
    iter_num = floor(log( b_to_the_iter_num )/log( b )); %(5) in articleelse
    w = 1+2*b.^(1:iter_num);
else
    iter_num = floor((w_max-1)/(2*b));
    w = 1+2*b.*(1:iter_num);
end

%% calculae dh_T (PCL: height_thresholds)
w_k = w(2:end);
w_kminus1 = w(1:end-1);

dh_T = [dh_0, dh_0+c*s.*(w_k-w_kminus1)];
dh_T(w_k<=3) = dh_0;
if do_global_thresholding
    dh_T(dh_T>dh_max) = dh_max;
end



%% estimate ground points
% initialize all as ground
ground_indices = 1:length(XY_indices);

flag = zeros(size(A)); % saves index of windows that brought to ground-classification (in article the window-size utself is saved)

for k = 1:iter_num
    A_f = imopen(A, strel('square',w(k))); % A_filterred
%    A_f = imopen(A, strel('line',w(k), 90)); % A_filterred
%    A_f = imopen(A_f, strel('line',w(k), 0)); % A_filterred
   is_ground = (A-A_f)>dh_T(k);
   flag(is_ground) = k;

    % classify points according to pixel values and current threshold
    %% MUST OPTIMIZE COMPLEXITY OF THIS! perhaps use linear indicies
    ground_bool = flag == 0;
    for row=1:length(ground_indices) 
        if ground_bool(XY_indices(row,2), XY_indices(row,1))
            if pc.Location(row, 3) > dh_T(k) + A_f(XY_indices(row,2), XY_indices(row,1))
                ground_indices(row) = inf;
            end
        else
            ground_indices(row) = inf;
        end
    end
    %%
   A = A_f;
end

ground_points_bool = not(isinf(ground_indices));

end

